@@ -43,7 +43,9 @@ enum
 	PROP_ICON,
 	PROP_PROPOSALS_BATCH_SIZE,
 	PROP_SCAN_BATCH_SIZE,
-	PROP_MINIMUM_WORD_SIZE
+	PROP_MINIMUM_WORD_SIZE,
+	PROP_INTERACTIVE_DELAY,
+	PROP_PRIORITY
 };
 
 struct _GtkSourceCompletionWordsPrivate
@@ -66,6 +68,9 @@ struct _GtkSourceCompletionWordsPrivate
 	
 	GtkSourceCompletionWordsLibrary *library;
 	GList *buffers;
+
+	gint interactive_delay;
+	gint priority;
 };
 
 typedef struct
@@ -89,10 +94,10 @@ G_DEFINE_TYPE_WITH_CODE (GtkSourceCompletionWords,
 			 G_IMPLEMENT_INTERFACE (GTK_TYPE_SOURCE_COMPLETION_PROVIDER,
 				 		gtk_source_completion_words_iface_init))
 
-static const gchar * 
+static gchar *
 gtk_source_completion_words_get_name (GtkSourceCompletionProvider *self)
 {
-	return GTK_SOURCE_COMPLETION_WORDS (self)->priv->name;
+	return g_strdup (GTK_SOURCE_COMPLETION_WORDS (self)->priv->name);
 }
 
 static GdkPixbuf * 
@@ -224,25 +229,12 @@ get_word_at_iter (GtkTextIter    *iter,
 		return gtk_text_iter_get_text (iter, &end);
 	}
 }
+
 static gboolean
 gtk_source_completion_words_match (GtkSourceCompletionProvider *provider,
                                    GtkSourceCompletionContext  *context)
 {
-	GtkSourceCompletionWords *words = GTK_SOURCE_COMPLETION_WORDS (provider);
-	GtkTextIter iter;
-	gchar *word;
-	gboolean ret;
-
-	gtk_source_completion_context_get_iter (context, &iter);
-	word = get_word_at_iter (&iter,
-	                         valid_word_char,
-	                         valid_start_char,
-	                         words);
-	
-	ret = word != NULL && g_utf8_strlen (word, -1) >= words->priv->minimum_word_size;
-	g_free (word);
-	
-	return ret;
+	return TRUE;
 }
 
 static void
@@ -265,8 +257,8 @@ gtk_source_completion_words_populate (GtkSourceCompletionProvider *provider,
 	                         valid_word_char,
 	                         valid_start_char,
 	                         words);
-	
-	if (word == NULL || 
+
+	if (word == NULL ||
 	    g_utf8_strlen (word, -1) < words->priv->minimum_word_size)
 	{
 		g_free (word);
@@ -406,6 +398,12 @@ gtk_source_completion_words_set_property (GObject      *object,
 			self->priv->minimum_word_size = g_value_get_uint (value);
 			update_buffers_minimum_word_size (self);
 		break;
+		case PROP_INTERACTIVE_DELAY:
+			self->priv->interactive_delay = g_value_get_int (value);
+		break;
+		case PROP_PRIORITY:
+			self->priv->priority = g_value_get_int (value);
+		break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -437,6 +435,12 @@ gtk_source_completion_words_get_property (GObject    *object,
 		case PROP_MINIMUM_WORD_SIZE:
 			g_value_set_uint (value, self->priv->minimum_word_size);
 		break;
+		case PROP_INTERACTIVE_DELAY:
+			g_value_set_int (value, self->priv->interactive_delay);
+		break;
+		case PROP_PRIORITY:
+			g_value_set_int (value, self->priv->priority);
+		break;
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -496,9 +500,29 @@ gtk_source_completion_words_class_init (GtkSourceCompletionWordsClass *klass)
 	                                                    _("The minimum word size to complete"),
 	                                                    2,
 	                                                    G_MAXUINT,
-	                                                    3,
+	                                                    2,
 	                                                    G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
-	
+
+	g_object_class_install_property (object_class,
+	                                 PROP_INTERACTIVE_DELAY,
+	                                 g_param_spec_int ("interactive-delay",
+	                                                   _("Interactive Delay"),
+	                                                   _("The delay before initiating interactive completion"),
+	                                                   -1,
+	                                                   G_MAXINT,
+	                                                   50,
+	                                                   G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+
+	g_object_class_install_property (object_class,
+	                                 PROP_PRIORITY,
+	                                 g_param_spec_int ("priority",
+	                                                   _("Priority"),
+	                                                   _("Provider priority"),
+	                                                   G_MININT,
+	                                                   G_MAXINT,
+	                                                   0,
+	                                                   G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+
 	g_type_class_add_private (object_class, sizeof(GtkSourceCompletionWordsPrivate));
 }
 
@@ -523,6 +547,18 @@ gtk_source_completion_words_get_start_iter (GtkSourceCompletionProvider *provide
 	return TRUE;
 }
 
+static gint
+gtk_source_completion_words_get_interactive_delay (GtkSourceCompletionProvider *provider)
+{
+	return GTK_SOURCE_COMPLETION_WORDS (provider)->priv->interactive_delay;
+}
+
+static gint
+gtk_source_completion_words_get_priority (GtkSourceCompletionProvider *provider)
+{
+	return GTK_SOURCE_COMPLETION_WORDS (provider)->priv->priority;
+}
+
 static void
 gtk_source_completion_words_iface_init (GtkSourceCompletionProviderIface *iface)
 {
@@ -533,6 +569,8 @@ gtk_source_completion_words_iface_init (GtkSourceCompletionProviderIface *iface)
 	iface->match = gtk_source_completion_words_match;
 
 	iface->get_start_iter = gtk_source_completion_words_get_start_iter;
+	iface->get_interactive_delay = gtk_source_completion_words_get_interactive_delay;
+	iface->get_priority = gtk_source_completion_words_get_priority;
 }
 
 static void 
