@@ -1,5 +1,5 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*-
- *  gtksourcebuffer.c
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8; coding: utf-8 -*- */
+/*  gtksourcebuffer.c
  *
  *  Copyright (C) 1999,2000,2001,2002 by:
  *          Mikael Hermansson <tyan@linux.se>
@@ -39,6 +39,26 @@
 #include "gtksourceiter.h"
 #include "gtksourcestyleschememanager.h"
 #include "gtksourcestyle-private.h"
+#include "gtksourceundomanagerdefault.h"
+
+/**
+ * SECTION:buffer
+ * @Short_description: Buffer object for #GtkSourceView
+ * @Title: GtkSourceBuffer
+ * @See_also: #GtkTextBuffer,#GtkSourceView
+ *
+ * The #GtkSourceBuffer object is the model for #GtkSourceView widgets.
+ * It extends the #GtkTextBuffer object by adding features useful to display
+ * and edit source code as syntax highlighting and bracket matching. It
+ * also implements support for undo/redo operations.
+ *
+ * To create a #GtkSourceBuffer use gtk_source_buffer_new() or
+ * gtk_source_buffer_new_with_language(). The second form is just a convenience
+ * function which allows you to initially set a #GtkSourceLanguage.
+ *
+ * By default highlighting is enabled, but you can disable it with
+ * gtk_source_buffer_set_highlight_syntax().
+ */
 
 /*
 #define ENABLE_DEBUG
@@ -61,10 +81,14 @@
 
 #define MAX_CHARS_BEFORE_FINDING_A_MATCH    10000
 
+#define TAG_CONTEXT_CLASS_NAME "GtkSourceViewTagContextClassName"
+
 /* Signals */
 enum {
 	HIGHLIGHT_UPDATED,
 	SOURCE_MARK_UPDATED,
+	UNDO,
+	REDO,
 	LAST_SIGNAL
 };
 
@@ -77,7 +101,8 @@ enum {
 	PROP_HIGHLIGHT_MATCHING_BRACKETS,
 	PROP_MAX_UNDO_LEVELS,
 	PROP_LANGUAGE,
-	PROP_STYLE_SCHEME
+	PROP_STYLE_SCHEME,
+	PROP_UNDO_MANAGER
 };
 
 struct _GtkSourceBufferPrivate
@@ -99,6 +124,7 @@ struct _GtkSourceBufferPrivate
 	GtkSourceStyleScheme  *style_scheme;
 
 	GtkSourceUndoManager  *undo_manager;
+	gint                   max_undo_levels;
 };
 
 G_DEFINE_TYPE (GtkSourceBuffer, gtk_source_buffer, GTK_TYPE_TEXT_BUFFER)
@@ -118,11 +144,9 @@ static void      gtk_source_buffer_get_property         (GObject
 							 guint                    prop_id,
 							 GValue                  *value,
 							 GParamSpec              *pspec);
-static void 	 gtk_source_buffer_can_undo_handler 	(GtkSourceUndoManager    *um,
-							 gboolean                 can_undo,
+static void 	 gtk_source_buffer_can_undo_handler 	(GtkSourceUndoManager    *manager,
 							 GtkSourceBuffer         *buffer);
-static void 	 gtk_source_buffer_can_redo_handler	(GtkSourceUndoManager    *um,
-							 gboolean                 can_redo,
+static void 	 gtk_source_buffer_can_redo_handler	(GtkSourceUndoManager    *manager,
 							 GtkSourceBuffer         *buffer);
 static void 	 gtk_source_buffer_real_insert_text 	(GtkTextBuffer           *buffer,
 							 GtkTextIter             *iter,
@@ -142,8 +166,12 @@ static void 	 gtk_source_buffer_real_mark_set	(GtkTextBuffer		 *buffer,
 							 GtkTextMark		 *mark);
 static void 	 gtk_source_buffer_real_mark_deleted	(GtkTextBuffer		 *buffer,
 							 GtkTextMark		 *mark);
-static gboolean	 gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter    *orig,
-								  gint            max_chars);
+static gboolean	 gtk_source_buffer_find_bracket_match_with_limit (GtkSourceBuffer *buffer,
+								  GtkTextIter     *orig,
+								  gint             max_chars);
+
+static void	 gtk_source_buffer_real_undo		(GtkSourceBuffer	 *buffer);
+static void	 gtk_source_buffer_real_redo		(GtkSourceBuffer	 *buffer);
 
 static void
 gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
@@ -169,6 +197,9 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	tb_class->mark_set	= gtk_source_buffer_real_mark_set;
 	tb_class->mark_deleted	= gtk_source_buffer_real_mark_deleted;
 
+	klass->undo = gtk_source_buffer_real_undo;
+	klass->redo = gtk_source_buffer_real_redo;
+
 	/**
 	 * GtkSourceBuffer:highlight-syntax:
 	 *
@@ -199,7 +230,8 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	/**
 	 * GtkSourceBuffer:max-undo-levels:
 	 *
-	 * Number of undo levels for the buffer. -1 means no limit.
+	 * Number of undo levels for the buffer. -1 means no limit. This property
+	 * will only affect the default undo manager.
 	 */
 	g_object_class_install_property (object_class,
 					 PROP_MAX_UNDO_LEVELS,
@@ -254,6 +286,14 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 							      GTK_TYPE_SOURCE_STYLE_SCHEME,
 							      G_PARAM_READWRITE));
 
+	g_object_class_install_property (object_class,
+	                                 PROP_UNDO_MANAGER,
+	                                 g_param_spec_object ("undo-manager",
+	                                                      _("Undo manager"),
+	                                                      _("The buffer undo manager"),
+	                                                      GTK_TYPE_SOURCE_UNDO_MANAGER,
+	                                                      G_PARAM_READWRITE | G_PARAM_CONSTRUCT));
+
 	param_types[0] = GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE;
 	param_types[1] = GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE;
 
@@ -283,9 +323,72 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			   G_TYPE_NONE,
 			   1, GTK_TYPE_TEXT_MARK);
 
+	buffer_signals[UNDO] =
+	    g_signal_new ("undo",
+			  G_OBJECT_CLASS_TYPE (object_class),
+			  G_SIGNAL_RUN_LAST,
+			  G_STRUCT_OFFSET (GtkSourceBufferClass, undo),
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
+			  G_TYPE_NONE,
+			  0);
+
+	buffer_signals[REDO] =
+	    g_signal_new ("redo",
+			  G_OBJECT_CLASS_TYPE (object_class),
+			  G_SIGNAL_RUN_LAST,
+			  G_STRUCT_OFFSET (GtkSourceBufferClass, redo),
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
+			  G_TYPE_NONE,
+			  0);
+
 	g_type_class_add_private (object_class, sizeof(GtkSourceBufferPrivate));
 }
 
+static void
+set_undo_manager (GtkSourceBuffer      *buffer,
+                  GtkSourceUndoManager *manager)
+{
+	if (manager == buffer->priv->undo_manager)
+	{
+		return;
+	}
+
+	if (buffer->priv->undo_manager != NULL)
+	{
+		g_signal_handlers_disconnect_by_func (buffer->priv->undo_manager,
+		                                      G_CALLBACK (gtk_source_buffer_can_undo_handler),
+		                                      buffer);
+
+		g_signal_handlers_disconnect_by_func (buffer->priv->undo_manager,
+		                                      G_CALLBACK (gtk_source_buffer_can_redo_handler),
+		                                      buffer);
+
+		g_object_unref (buffer->priv->undo_manager);
+		buffer->priv->undo_manager = NULL;
+	}
+
+	if (manager != NULL)
+	{
+		buffer->priv->undo_manager = g_object_ref (manager);
+
+		g_signal_connect (buffer->priv->undo_manager,
+		                  "can-undo-changed",
+		                  G_CALLBACK (gtk_source_buffer_can_undo_handler),
+		                  buffer);
+
+		g_signal_connect (buffer->priv->undo_manager,
+		                  "can-redo-changed",
+		                  G_CALLBACK (gtk_source_buffer_can_redo_handler),
+		                  buffer);
+
+		/* Notify possible changes in the can-undo/redo state */
+		g_object_notify (G_OBJECT (buffer), "can-undo");
+		g_object_notify (G_OBJECT (buffer), "can-redo");
+	}
+}
+
 static void
 gtk_source_buffer_init (GtkSourceBuffer *buffer)
 {
@@ -296,27 +399,16 @@ gtk_source_buffer_init (GtkSourceBuffer *buffer)
 
 	buffer->priv = priv;
 
-	priv->undo_manager = gtk_source_undo_manager_new (GTK_TEXT_BUFFER (buffer));
-
 	priv->highlight_syntax = TRUE;
 	priv->highlight_brackets = TRUE;
 	priv->bracket_mark = NULL;
 	priv->bracket_found = FALSE;
 
 	priv->source_marks = g_array_new (FALSE, FALSE, sizeof (GtkSourceMark *));
-
 	priv->style_scheme = _gtk_source_style_scheme_get_default ();
+
 	if (priv->style_scheme != NULL)
 		g_object_ref (priv->style_scheme);
-
-	g_signal_connect (priv->undo_manager,
-			  "can_undo",
-			  G_CALLBACK (gtk_source_buffer_can_undo_handler),
-			  buffer);
-	g_signal_connect (priv->undo_manager,
-			  "can_redo",
-			  G_CALLBACK (gtk_source_buffer_can_redo_handler),
-			  buffer);
 }
 
 static GObject *
@@ -325,13 +417,21 @@ gtk_source_buffer_constructor (GType                  type,
 			       GObjectConstructParam *construct_param)
 {
 	GObject *object;
+	GtkSourceBuffer *buffer;
 
 	object = G_OBJECT_CLASS(gtk_source_buffer_parent_class)->constructor (type,
 									      n_construct_properties,
 									      construct_param);
 
 	/* we need to know that the tag-table was set */
-	GTK_SOURCE_BUFFER (object)->priv->constructed = TRUE;
+	buffer = GTK_SOURCE_BUFFER (object);
+	buffer->priv->constructed = TRUE;
+
+	if (buffer->priv->undo_manager == NULL)
+	{
+		/* This will install the default undo manager */
+		gtk_source_buffer_set_undo_manager (buffer, NULL);
+	}
 
 	return object;
 }
@@ -366,8 +466,7 @@ gtk_source_buffer_dispose (GObject *object)
 
 	if (buffer->priv->undo_manager != NULL)
 	{
-		g_object_unref (buffer->priv->undo_manager);
-		buffer->priv->undo_manager = NULL;
+		set_undo_manager (buffer, NULL);
 	}
 
 	if (buffer->priv->highlight_engine != NULL)
@@ -431,6 +530,11 @@ gtk_source_buffer_set_property (GObject      *object,
 							    g_value_get_object (value));
 			break;
 
+		case PROP_UNDO_MANAGER:
+			gtk_source_buffer_set_undo_manager (source_buffer,
+			                                    g_value_get_object (value));
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -463,7 +567,7 @@ gtk_source_buffer_get_property (GObject    *object,
 
 		case PROP_MAX_UNDO_LEVELS:
 			g_value_set_int (value,
-					 gtk_source_buffer_get_max_undo_levels (source_buffer));
+					 source_buffer->priv->max_undo_levels);
 			break;
 
 		case PROP_LANGUAGE:
@@ -482,6 +586,10 @@ gtk_source_buffer_get_property (GObject    *object,
 			g_value_set_boolean (value, gtk_source_buffer_can_redo (source_buffer));
 			break;
 
+		case PROP_UNDO_MANAGER:
+			g_value_set_object (value, source_buffer->priv->undo_manager);
+			break;
+
 		default:
 			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 			break;
@@ -530,9 +638,8 @@ gtk_source_buffer_new_with_language (GtkSourceLanguage *language)
 }
 
 static void
-gtk_source_buffer_can_undo_handler (GtkSourceUndoManager  	*um,
-				    gboolean			 can_undo,
-				    GtkSourceBuffer 		*buffer)
+gtk_source_buffer_can_undo_handler (GtkSourceUndoManager *manager,
+                                    GtkSourceBuffer      *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
 
@@ -540,9 +647,8 @@ gtk_source_buffer_can_undo_handler (GtkSourceUndoManager  	*um,
 }
 
 static void
-gtk_source_buffer_can_redo_handler (GtkSourceUndoManager  	*um,
-				    gboolean         		 can_redo,
-				    GtkSourceBuffer 		*buffer)
+gtk_source_buffer_can_redo_handler (GtkSourceUndoManager *manager,
+                                    GtkSourceBuffer      *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
 
@@ -578,6 +684,17 @@ get_bracket_match_tag (GtkSourceBuffer *buffer)
 	return buffer->priv->bracket_match_tag;
 }
 
+/*
+ * This is private, just used by the compositor to not print bracket
+ * matches. Note that unlike get_bracket_match_tag() it returns NULL
+ * if the tag is not set.
+ */
+GtkTextTag *
+_gtk_source_buffer_get_bracket_match_tag (GtkSourceBuffer *buffer)
+{
+	return buffer->priv->bracket_match_tag;
+}
+
 static void
 gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 			       const GtkTextIter *iter,
@@ -610,7 +727,9 @@ gtk_source_buffer_move_cursor (GtkTextBuffer     *buffer,
 		return;
 
 	iter1 = *iter;
-	if (gtk_source_buffer_find_bracket_match_with_limit (&iter1, MAX_CHARS_BEFORE_FINDING_A_MATCH))
+	if (gtk_source_buffer_find_bracket_match_with_limit (GTK_SOURCE_BUFFER (buffer),
+	                                                     &iter1,
+	                                                     MAX_CHARS_BEFORE_FINDING_A_MATCH))
 	{
 		if (!GTK_SOURCE_BUFFER (buffer)->priv->bracket_mark)
 			GTK_SOURCE_BUFFER (buffer)->priv->bracket_mark =
@@ -771,35 +890,36 @@ gtk_source_buffer_real_delete_range (GtkTextBuffer *buffer,
 						 offset, length);
 }
 
-/* FIXME: this can't be here, but it's now needed for bracket matching */
-static gpointer
-iter_has_syntax_tag (const GtkTextIter *iter)
+/* This describes a mask of relevant context classes for highlighting matching
+   brackets. Additional classes can be added below */
+static const gchar *cclass_mask_definitions[] = {
+	"comment",
+	"string",
+};
+
+static gint
+get_context_class_mask (GtkSourceBuffer *buffer,
+                        GtkTextIter     *iter)
 {
-	return NULL;
-// 	const GtkSourceTag *tag;
-// 	GSList *list;
-// 	GSList *l;
-//
-// 	g_return_val_if_fail (iter != NULL, NULL);
-//
-// 	list = gtk_text_iter_get_tags (iter);
-// 	tag = NULL;
-//
-// 	l = list;
-//
-// 	while ((list != NULL) && (tag == NULL)) {
-// 		if (GTK_IS_SOURCE_TAG (list->data))
-// 			tag = GTK_SOURCE_TAG (list->data);
-// 		list = g_slist_next (list);
-// 	}
-//
-// 	g_slist_free (l);
-//
-// 	return tag;
+	gint i;
+	gint ret = 0;
+
+	for (i = 0; i < sizeof (cclass_mask_definitions) / sizeof (gchar *); ++i)
+	{
+		gboolean hasclass = gtk_source_buffer_iter_has_context_class (buffer,
+		                                                              iter,
+		                                                              cclass_mask_definitions[i]);
+
+		ret |= hasclass << i;
+	}
+
+	return ret;
 }
 
 static gboolean
-gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
+gtk_source_buffer_find_bracket_match_real (GtkSourceBuffer *buffer,
+                                           GtkTextIter     *orig,
+                                           gint             max_chars)
 {
 	GtkTextIter iter;
 
@@ -812,14 +932,14 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
 
 	gboolean found;
 
-	gpointer base_tag;
+	gint cclass_mask;
 
 	iter = *orig;
 
 	cur_char = gtk_text_iter_get_char (&iter);
 
 	base_char = search_char = cur_char;
-	base_tag = iter_has_syntax_tag (&iter);
+	cclass_mask = get_context_class_mask (buffer, &iter);
 
 	switch ((int) base_char) {
 		case '{':
@@ -867,12 +987,24 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
 	char_cont = 0;
 
 	do {
+		gint current_mask;
+
 		gtk_text_iter_forward_chars (&iter, addition);
 		cur_char = gtk_text_iter_get_char (&iter);
 		++char_cont;
 
+		current_mask = get_context_class_mask (buffer, &iter);
+
+		/* Check if we lost a class, which means we don't look any
+		   further */
+		if (current_mask < cclass_mask)
+		{
+			found = FALSE;
+			break;
+		}
+
 		if ((cur_char == search_char || cur_char == base_char) &&
-		    base_tag == iter_has_syntax_tag (&iter))
+		    cclass_mask == current_mask)
 		{
 			if ((cur_char == search_char) && counter == 0) {
 				found = TRUE;
@@ -898,11 +1030,13 @@ gtk_source_buffer_find_bracket_match_real (GtkTextIter *orig, gint max_chars)
  * cursor takes precedence.
  */
 static gboolean
-gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter *orig, gint max_chars)
+gtk_source_buffer_find_bracket_match_with_limit (GtkSourceBuffer *buffer,
+                                                 GtkTextIter     *orig,
+                                                 gint             max_chars)
 {
 	GtkTextIter iter;
 
-	if (gtk_source_buffer_find_bracket_match_real (orig, max_chars))
+	if (gtk_source_buffer_find_bracket_match_real (buffer, orig, max_chars))
 	{
 		return TRUE;
 	}
@@ -911,7 +1045,7 @@ gtk_source_buffer_find_bracket_match_with_limit (GtkTextIter *orig, gint max_cha
 	if (!gtk_text_iter_starts_line (&iter) &&
 	    gtk_text_iter_backward_char (&iter))
 	{
-		if (gtk_source_buffer_find_bracket_match_real (&iter, max_chars))
+		if (gtk_source_buffer_find_bracket_match_real (buffer, &iter, max_chars))
 		{
 			*orig = iter;
 			return TRUE;
@@ -994,9 +1128,8 @@ void
 gtk_source_buffer_undo (GtkSourceBuffer *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (gtk_source_undo_manager_can_undo (buffer->priv->undo_manager));
 
-	gtk_source_undo_manager_undo (buffer->priv->undo_manager);
+	g_signal_emit (buffer, buffer_signals[UNDO], 0);
 }
 
 /**
@@ -1010,9 +1143,8 @@ void
 gtk_source_buffer_redo (GtkSourceBuffer *buffer)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (gtk_source_undo_manager_can_redo (buffer->priv->undo_manager));
 
-	gtk_source_undo_manager_redo (buffer->priv->undo_manager);
+	g_signal_emit (buffer, buffer_signals[REDO], 0);
 }
 
 /**
@@ -1030,7 +1162,7 @@ gtk_source_buffer_get_max_undo_levels (GtkSourceBuffer *buffer)
 {
 	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), 0);
 
-	return gtk_source_undo_manager_get_max_undo_levels (buffer->priv->undo_manager);
+	return buffer->priv->max_undo_levels;
 }
 
 /**
@@ -1057,13 +1189,20 @@ gtk_source_buffer_set_max_undo_levels (GtkSourceBuffer *buffer,
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
 
-	if (gtk_source_undo_manager_get_max_undo_levels (
-		    buffer->priv->undo_manager) != max_undo_levels)
+	if (buffer->priv->max_undo_levels == max_undo_levels)
 	{
-		gtk_source_undo_manager_set_max_undo_levels (buffer->priv->undo_manager,
-							     max_undo_levels);
-		g_object_notify (G_OBJECT (buffer), "max-undo-levels");
+		return;
 	}
+
+	buffer->priv->max_undo_levels = max_undo_levels;
+
+	if (GTK_IS_SOURCE_UNDO_MANAGER_DEFAULT (buffer->priv->undo_manager))
+	{
+		gtk_source_undo_manager_default_set_max_undo_levels (GTK_SOURCE_UNDO_MANAGER_DEFAULT (buffer->priv->undo_manager),
+		                                                     max_undo_levels);
+	}
+
+	g_object_notify (G_OBJECT (buffer), "max-undo-levels");
 }
 
 /**
@@ -1225,6 +1364,7 @@ gtk_source_buffer_set_language (GtkSourceBuffer   *buffer,
 				GtkSourceLanguage *language)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
+	g_return_if_fail (GTK_IS_SOURCE_LANGUAGE (language) || language == NULL);
 
 	if (buffer->priv->language == language)
 		return;
@@ -1337,14 +1477,15 @@ gtk_source_buffer_ensure_highlight (GtkSourceBuffer   *buffer,
  * @buffer: a #GtkSourceBuffer.
  * @scheme: style scheme.
  *
- * Sets style scheme used by the buffer.
+ * Sets style scheme used by the buffer. If @scheme is %NULL no
+ * style scheme is used.
  **/
 void
 gtk_source_buffer_set_style_scheme (GtkSourceBuffer      *buffer,
 				    GtkSourceStyleScheme *scheme)
 {
 	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
-	g_return_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme));
+	g_return_if_fail (GTK_IS_SOURCE_STYLE_SCHEME (scheme) || scheme == NULL);
 
 	if (buffer->priv->style_scheme == scheme)
 		return;
@@ -1352,7 +1493,7 @@ gtk_source_buffer_set_style_scheme (GtkSourceBuffer      *buffer,
 	if (buffer->priv->style_scheme)
 		g_object_unref (buffer->priv->style_scheme);
 
-	buffer->priv->style_scheme = g_object_ref (scheme);
+	buffer->priv->style_scheme = scheme ? g_object_ref (scheme) : NULL;
 	update_bracket_match_style (buffer);
 
 	if (buffer->priv->highlight_engine != NULL)
@@ -1524,6 +1665,22 @@ gtk_source_buffer_real_mark_deleted (GtkTextBuffer *buffer,
 		GTK_TEXT_BUFFER_CLASS (gtk_source_buffer_parent_class)->mark_deleted (buffer, mark);
 }
 
+static void
+gtk_source_buffer_real_undo (GtkSourceBuffer *buffer)
+{
+	g_return_if_fail (gtk_source_undo_manager_can_undo (buffer->priv->undo_manager));
+
+	gtk_source_undo_manager_undo (buffer->priv->undo_manager);
+}
+
+static void
+gtk_source_buffer_real_redo (GtkSourceBuffer *buffer)
+{
+	g_return_if_fail (gtk_source_undo_manager_can_redo (buffer->priv->undo_manager));
+
+	gtk_source_undo_manager_redo (buffer->priv->undo_manager);
+}
+
 /**
  * gtk_source_buffer_create_source_mark:
  * @buffer: a #GtkSourceBuffer.
@@ -1929,3 +2086,224 @@ gtk_source_buffer_remove_source_marks (GtkSourceBuffer   *buffer,
 	g_slist_free (list);
 }
 
+
+/**
+ * gtk_source_buffer_iter_has_context_class:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: class to search for
+ *
+ * Check if the class @context_klass is set on @iter.
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_has_context_class (GtkSourceBuffer   *buffer,
+                                          const GtkTextIter *iter,
+                                          const gchar       *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag != NULL)
+	{
+		return gtk_text_iter_has_tag (iter, tag);
+	}
+	else
+	{
+		return FALSE;
+	}
+}
+
+/**
+ * gtk_source_buffer_get_context_classes_at_iter:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ *
+ * Get all defined context classes at @iter.
+ *
+ * Returns: a new %NULL terminated array of context class names. Use
+ *          #g_strfreev to free the array if it is no longer needed.
+ *
+ * Since: 2.10
+ **/
+gchar **
+gtk_source_buffer_get_context_classes_at_iter (GtkSourceBuffer   *buffer,
+                                               const GtkTextIter *iter)
+{
+	GSList *tags;
+	GSList *item;
+	GPtrArray *ret;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
+	g_return_val_if_fail (iter != NULL, NULL);
+
+	tags = gtk_text_iter_get_tags (iter);
+	ret = g_ptr_array_new ();
+
+	for (item = tags; item; item = g_slist_next (item))
+	{
+		gchar const *name = g_object_get_data (G_OBJECT (item->data),
+		                                       TAG_CONTEXT_CLASS_NAME);
+
+		if (name != NULL)
+		{
+			g_ptr_array_add (ret, g_strdup (name));
+		}
+	}
+
+	g_ptr_array_add (ret, NULL);
+	return (gchar **) g_ptr_array_free (ret, FALSE);
+}
+
+/**
+ * gtk_source_buffer_iter_forward_to_context_class_toggle:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: the context class
+ *
+ * Moves forward to the next toggle (on or off) of the context class. If no
+ * matching context class toggles are found, returns %FALSE, otherwise %TRUE.
+ * Does not return toggles located at @iter, only toggles after @iter. Sets
+ * @iter to the location of the toggle, or to the end of the buffer if no
+ * toggle is found.
+ *
+ * Returns: whether we found a context class toggle after @iter
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_forward_to_context_class_toggle (GtkSourceBuffer *buffer,
+                                                        GtkTextIter     *iter,
+                                                        const gchar     *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag == NULL)
+	{
+		return FALSE;
+	}
+	else
+	{
+		return gtk_text_iter_forward_to_tag_toggle (iter, tag);
+	}
+}
+
+/**
+ * gtk_source_buffer_iter_backward_to_context_class_toggle:
+ * @buffer: a #GtkSourceBuffer.
+ * @iter: a #GtkTextIter
+ * @context_class: the context class
+ *
+ * Moves backward to the next toggle (on or off) of the context class. If no
+ * matching context class toggles are found, returns %FALSE, otherwise %TRUE.
+ * Does not return toggles located at @iter, only toggles after @iter. Sets
+ * @iter to the location of the toggle, or to the end of the buffer if no
+ * toggle is found.
+ *
+ * Returns: whether we found a context class toggle before @iter
+ *
+ * Since: 2.10
+ **/
+gboolean
+gtk_source_buffer_iter_backward_to_context_class_toggle (GtkSourceBuffer *buffer,
+                                                         GtkTextIter     *iter,
+                                                         const gchar     *context_class)
+{
+	GtkTextTag *tag;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), FALSE);
+	g_return_val_if_fail (iter != NULL, FALSE);
+	g_return_val_if_fail (context_class != NULL, FALSE);
+
+	if (buffer->priv->highlight_engine == NULL)
+	{
+		return FALSE;
+	}
+
+	tag = _gtk_source_engine_get_context_class_tag (buffer->priv->highlight_engine,
+							context_class);
+
+	if (tag == NULL)
+	{
+		return FALSE;
+	}
+	else
+	{
+		return gtk_text_iter_backward_to_tag_toggle (iter, tag);
+	}
+}
+
+/**
+ * gtk_source_buffer_set_undo_manager:
+ * @buffer: A #GtkSourceBuffer
+ * @manager: A #GtkSourceUndoManager
+ *
+ * Set the buffer undo manager. If @manager is %NULL the default undo manager
+ * will be set.
+ *
+ **/
+void
+gtk_source_buffer_set_undo_manager (GtkSourceBuffer      *buffer,
+                                    GtkSourceUndoManager *manager)
+{
+	g_return_if_fail (GTK_IS_SOURCE_BUFFER (buffer));
+	g_return_if_fail (manager == NULL || GTK_IS_SOURCE_UNDO_MANAGER (manager));
+
+	if (manager == NULL)
+	{
+		manager = g_object_new (GTK_TYPE_SOURCE_UNDO_MANAGER_DEFAULT,
+		                        "buffer", buffer,
+		                        "max-undo-levels", buffer->priv->max_undo_levels,
+		                        NULL);
+	}
+	else
+	{
+		g_object_ref (manager);
+	}
+
+	set_undo_manager (buffer, manager);
+	g_object_unref (manager);
+
+	g_object_notify (G_OBJECT (buffer), "undo-manager");
+}
+
+/**
+ * gtk_source_buffer_get_undo_manager:
+ * @buffer: A #GtkSourceBuffer
+ *
+ * Get the undo manager associated with the buffer.
+ *
+ * Returns: A #GtkSourceUndoManager
+ *
+ **/
+GtkSourceUndoManager *
+gtk_source_buffer_get_undo_manager (GtkSourceBuffer *buffer)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_BUFFER (buffer), NULL);
+
+	return buffer->priv->undo_manager;
+}
